import os
import sys

import gsdl2
import pytmxloader

import spatialhash
from sprite import TextureSprite
import textureutil


default_map_dir = 'data/map'

_texture_cache = {}
_surface_cache = {}


class BasicMap(object):
    def __init__(self, map_name='BasicMap', file_name=None, file_obj=None, map_data=None,
                 map_dir=default_map_dir, use_textures=True, object_factory=None):
        """
        Usage:
            BasicMap('jim', file_name='jimsmap.json')
            BasicMap('joe', file_obj=joe_fh)
            BasicMap('jack', map_data=json_data)

        object_factory's interface must conform to: object_factory(obj, rects, ellipses, polygons, poly_lines, tiles).

        :param map_name:
        :param file_name:
        :param file_obj:
        :param map_data:
        :param map_dir:
        :param use_textures: bool; True=create texture sprites; False=create legacy surface sprites
        :param object_factory: callable; create a game object from an instance of pytmxloader Object*
        """
        use_textures = bool(use_textures)
        self._use_textures = use_textures
        self._map_dir = map_dir
        self.object_factory = object_factory

        # Load the map info
        self.map_name = map_name
        if map_data:
            self.loaded_from = 'map_data'
            self.map_info = pytmxloader.pytmxloader.load_map_from_data(map_data, '')
        elif file_obj:
            self.loaded_from = 'file_like'
            self.map_info = pytmxloader.pytmxloader.load_map_from_file_like(file_obj)
        elif file_name:
            self.loaded_from = 'file_path: {}'.format('/'.join((map_dir, file_name)))
            self.map_info = pytmxloader.pytmxloader.load_map_from_file_path('/'.join((map_dir, file_name)))
        else:
            raise ValueError('no valid map data')

        # Calculate a "world" rect from the layers
        def getkeyw(o):
            return o.x + o.width
        def getkeyh(o):
            return o.y + o.height
        x = y = sys.maxint
        w = h = 1
        for layer in self.map_info.layers:
            x = min(x, layer.pixel_offset_x)
            y = min(y, layer.pixel_offset_y)
            if isinstance(layer, pytmxloader.TileLayerInfo):
                w = max(w, layer.width * layer.tile_w)
                h = max(h, layer.height * layer.tile_h)
            elif isinstance(layer, (pytmxloader.ImageLayerInfo, pytmxloader.ObjectLayerInfo)):
                maxobj = max(layer.objects, key=getkeyw)
                if maxobj:
                    w = max(w, getkeyw(maxobj))
                maxobj = max(layer.objects, key=getkeyh)
                if maxobj:
                    h = max(h, getkeyh(maxobj))
        self.rect = self.map_info.rect = gsdl2.Rect(x, y, w, h)

        if use_textures:
            load_textures(self.map_info, self._map_dir)
            self.layers = create_layers(self.map_info, use_textures=True, object_factory=object_factory)
        else:
            load_surfaces(self.map_info, self._map_dir)
            self.layers = create_layers(self.map_info, use_textures=False, object_factory=object_factory)

        self.layers_by_name = {}
        for layer in self.layers:
            self.layers_by_name[layer.name] = layer


class BasicLayer(spatialhash.SpatialHash):
    def __init__(self, rect, cell_size, raw_layer):
        super(BasicLayer, self).__init__(rect, cell_size)
        self.raw_layer = raw_layer
        self.name = raw_layer.name
        self.layer_type = raw_layer.layer_type
        self.visible = raw_layer.visible


def load_textures(map_info, map_dir=default_map_dir, img_cache=None):
    if img_cache is None:
        img_cache = _texture_cache
    for tile in map_info.tiles.values():
        tileset = tile.tileset
        tileset_name = tileset.name

        # load source image first if not already cached
        image_source = img_cache.get(tileset_name, None)
        if image_source is None:
            # print('map_dir={} image_dir={} name={}'.format(map_dir, tileset.image_path_rel_to_map, tileset_name))
            image_file_name = '/'.join((map_dir, tileset.image_path_rel_to_map))
            assert os.access(image_file_name, os.F_OK), 'file not found: {}'.format(image_file_name)
            image_source = gsdl2.image.load_texture(image_file_name)
            img_cache[tileset_name] = image_source

        # load tile image from source image
        key = (tileset_name, tile.spritesheet_x, tile.spritesheet_y)
        tile_image = img_cache.get(key, None)
        if tile_image is None:
            src_rect = gsdl2.Rect(tile.spritesheet_x, tile.spritesheet_y, tileset.tile_width, tileset.tile_height)
            tile_image = textureutil.copy(image_source, src_rect)
            img_cache[key] = tile_image
        tile.image = tile_image


def load_surfaces(map_info, map_dir=default_map_dir, img_cache=None):
    if img_cache is None:
        img_cache = _surface_cache
    for tile in map_info.tiles.values():
        tileset = tile.tileset
        tileset_name = tileset.name

        # load source image first if not already cached
        image_source = img_cache.get(tileset_name, None)
        if image_source is None:
            image_file_name = '/'.join((map_dir, tileset.image_path_rel_to_map))
            assert os.access(image_file_name, os.F_OK), 'file not found: {}'.format(image_file_name)
            image_source = gsdl2.image.load(image_file_name)
            if tileset.transparent_color:
                transparent_color = gsdl2.Color(tileset.transparent_color)
                image_source.set_colorkey(transparent_color, gsdl2.RLEACCEL)
                image_source = image_source.convert()
            else:
                image_source = image_source.convert_alpha()
            img_cache[tileset_name] = image_source

        # load tile image from source image
        key = (tileset_name, tile.spritesheet_x, tile.spritesheet_y, tile.angle, tile.flip_x, tile.flip_y)
        tile_image = img_cache.get(key, None)
        if tile_image is None:
            tile_image = gsdl2.Surface((tileset.tile_width, tileset.tile_height), gsdl2.SRCALPHA)
            # tile_image.fill((0, 0, 0, 0), None, pygame.BLEND_RGBA_MULT)
            source_rect = gsdl2.Rect(tile.spritesheet_x, tile.spritesheet_y, tileset.tile_width, tileset.tile_height)
            tile_image.blit(image_source, (0, 0), source_rect)
            # important: order of operations matter: rotate first and then flip the image
            # if tile.angle > 0:
            #     tile_image = pygame.transform.rotate(tile_image, -tile.angle)
            # if tile.flip_x or tile.flip_y:
            #     tile_image = pygame.transform.flip(tile_image, tile.flip_x, tile.flip_y)
            img_cache[key] = tile_image
        tile.image = tile_image


def create_layers(map_info, use_textures=True, object_factory=None):
    """Create list of friendly BasicLayer containing sprites and objects

    If use_textures is True then TextureSprites will be created; else legacy surface Sprites. (TODO)

    object_factory's interface must conform to: object_factory(obj, rects, ellipses, polygons, poly_lines, tiles).

    :param map_info: the MapInfo object from pytmxloader
    :param use_textures: bool; True=texture sprites; False=legacy surface sprites
    :param object_factory: callable; create a game object from an instance of pytmxloader Object*
    :return: list of BasicLayer
    """
    layers = []
    for raw_layer in map_info.layers:
        layer = BasicLayer(map_info.rect, 128, raw_layer)
        layers.append(layer)
        if isinstance(raw_layer, (pytmxloader.TileLayerInfo, pytmxloader.ImageLayerInfo)):
            for row in raw_layer.data_yx:
                for lti in row:
                    if lti is not None:
                        if use_textures:
                            sprite = TextureSprite(
                                lti.tile_info.image, (lti.tile_x, lti.tile_y + lti.offset_y), anchor='topleft')
                        else:
                            sprite = SurfaceSprite()
                        layer.add(sprite)
        elif isinstance(raw_layer, pytmxloader.ObjectLayerInfo) and object_factory is not None:
            object_factory(
                layer, raw_layer.rectangles, raw_layer.ellipses, raw_layer.polygons, raw_layer.poly_lines,
                raw_layer.tiles)
    return layers
