"""game.py - a context for running a game session or level
"""

import random

import gsdl2
from gsdl2.locals import S_ESCAPE

import context
import sound
import settings
import spatialhash
import sprite
import text
import tweening


tweener = tweening.Tweener()


def get_layer(obj):
    """sort key for sprite.TextureSprite in layer order

    See Game.draw.
    """
    return obj.layer


class Game(context.Context):
    """the game context"""

    def __init__(self):
        """initialize the game - all configurable items are in settings.py"""
        # init parent class
        context.Context.__init__(self, gsdl2.time.FixedDriver(self.update, settings.time_step))

        # create some schedules
        self.clock.new_schedule(self.set_caption, 0.1)
        self.clock.new_schedule(self.new_floating_text, 0.1)
        self.render_sched = self.clock.new_schedule(self.draw, 1.0 / settings.max_fps, keep_history=True)

        # cache renderer info for convenience - it is not going to change
        self.renderer = gsdl2.display.get_renderer()
        self.renderer.draw_color = 0, 0, 0

        # set music end event, and let self.on_user_event drive the music loop
        gsdl2.music.set_endevent(code=settings.song_finished)

        # Space is a SpatialHash, generally the best combination of collection and spatial partitioning algorithms
        # for culling and collision detection.
        # NOTE: whenever an object's position changes, it must be updated via settings.space.add() to recompute the
        # buckets. If this is not done, you will get strange results such as: some collisions wrongly detected; some
        # objects not displayed.
        settings.space = spatialhash.SpatialHash(settings.screen_rect.inflate(100, 100), 64)

    def set_caption(self, dt):
        """scheduled callback to update the caption"""
        try:
            gsdl2.display.set_caption('{} fps | Now playing: {}'.format(
                int(self.render_sched.per_second()), settings.music_now_playing))
        except ZeroDivisionError:
            # this handler was inspired by variables that can occasionally be zero :)
            pass

    def new_floating_text(self, dt):
        """Remove me, I'm part of the demo - scheduled callback to create some floating texts"""
        x = random.randrange(settings.screen_rect.right)
        y = random.randrange(settings.screen_rect.bottom)
        text.make_floating_score(
            '{:d} fps'.format(self.render_sched.per_second()), (x, y), layer=settings.layer_floating_text)

    def enter(self):
        """called when the this Context is pushed onto the stack

        See also: context.Context.
        """
        gsdl2.event.set_grab(settings.ARG_GRAB, gsdl2.display.get_window())
        gsdl2.mouse.set_visible(False)
        sound.update_music()

    def resume(self):
        """called when this Context is resumed; this occurs when the higher Context is popped off the stack

        See also: context.Context.
        """
        gsdl2.event.set_grab(settings.ARG_GRAB, gsdl2.display.get_window())
        gsdl2.mouse.set_visible(False)
        sound.update_music()

    def suspend(self):
        """called when this Context is suspended; this occurs when another Context is pushed on the stack

        See also: context.Context.
        """
        gsdl2.event.set_grab(False, gsdl2.display.get_window())
        gsdl2.mouse.set_visible(True)

    def exit(self):
        """called when this Context is popped off the stack

        See also: context.Context.
        """
        gsdl2.event.set_grab(False, gsdl2.display.get_window())
        gsdl2.mouse.set_visible(True)
        gsdl2.display.set_caption(settings.game_title)
        # Fade out if music is not the intro music (assumed index 0). Customize if you don't like this behavior.
        if settings.music_now_playing != settings.music_info[0]['name']:
            gsdl2.music.fadeout(1000)

    def update(self, dt):
        """scheduled callback that updates the game logic

        This is called on a fixed step timer. The clock tries to keep accurate subjective time, but even if CPU lags
        it will always provide the same dt.

        You probably want to leave this as it is. You can add things, but don't remove them unless you are sure you
        won't use them.

        See also: Game.__init__, context.Context, tweeening.Tweener, text module, sprite module.
        """
        self.get_events()
        tweener.update(dt)
        self.update_game(dt)
        text.update(dt)
        sprite.update(dt)

    def update_game(self, dt):
        """Add your game logic here"""
        # Add some game logic...
        pass

    def draw(self, sched):
        """scheduled callback that draws all the sprites in settings.space

        See also: Game.__init__, renderer.Renderer, settings.max_fps, display module.
        """
        interp = sched.interp
        renderer = self.renderer
        renderer.clear()
        for obj in sorted(settings.space, key=get_layer):
            obj.draw(renderer, interp)
        renderer.present()

    def on_key_down(self, e):
        """overrides Context.on_key_down"""
        if e.scancode == S_ESCAPE:
            self.on_quit()

    def on_user_event(self, e):
        """overrides Context.on_user_event"""
        if e.code == settings.song_finished:
            sound.update_music()

    def on_quit(self):
        """overrides Context.on_quit"""
        self.pop()
