"""tiledutils.py - utils for making game objects from Tiled objects, and rendering them
"""

import sys

import gsdl2
from gsdl2 import gfx

import geometry
from vec2d import Vec2d


# Object sprites pick up the default color at the time of their creation.
# obj.properties["color"] override this default.
default_color = 'white'


class Rectangle(geometry.RectGeometry):
    def __init__(self, obj, position=None, anchor='topleft'):
        super(Rectangle, self). __init__(obj.x, obj.y, obj.width, obj.height, position=position, anchor=anchor)
        self.raw_obj = obj
        self.name = obj.name
        self.properties = obj.properties
        color = obj.properties.get('color', default_color).encode('utf-8')
        self.color = gsdl2.Color(color)
        self.visible = obj.visible


class Ellipse(geometry.EllipseGeometry):
    def __init__(self, obj):
        super(Ellipse, self). __init__((obj.x, obj.y), obj.width / 2.0, obj.height / 2.0)
        self.raw_obj = obj
        self.name = obj.name
        self.properties = obj.properties
        color = obj.properties.get('color', default_color).encode('utf-8')
        self.color = gsdl2.Color(color)
        self.visible = obj.visible
        self.position = self.rect.bottomright


class Circle(geometry.CircleGeometry):
    def __init__(self, obj):
        super(Circle, self). __init__((obj.x, obj.y), obj.width / 2.0)
        self.raw_obj = obj
        self.name = obj.name
        self.properties = obj.properties
        color = obj.properties.get('color', default_color).encode('utf-8')
        self.color = gsdl2.Color(color)
        self.visible = obj.visible
        self.position = self.rect.bottomright


class Polygon(geometry.PolyGeometry):
    def __init__(self, obj, closed=True, anchor='topleft'):
        x, y, points = self.poly_translate(obj)
        super(Polygon, self). __init__(points, position=(x, y), closed=closed, anchor=anchor)
        self.raw_obj = obj
        self.name = obj.name
        self.properties = obj.properties
        color = obj.properties.get('color', default_color).encode('utf-8')
        self.color = gsdl2.Color(color)
        self.visible = obj.visible

    @staticmethod
    def poly_translate(obj):
        def min_seq(seq):
            n = sys.maxint
            for s in seq:
                if s < n:
                    n = s
            return n
        pixel_points = obj.pixel_points
        minx = min_seq([p[0] for p in pixel_points])
        miny = min_seq([p[1] for p in pixel_points])
        d = minx, miny
        points = [Vec2d(p) - d for p in pixel_points]
        return minx, miny, points


def object_factory(layer, rectangles, ellipses, polygons, poly_lines, tiles):
    """create sprites from Tiled objects

    This is a default factory for use with basicmap.BasicMap.

    To leverage this factory in a custom factory, create the custom factory and handle the custom objects, and add a
    case to call this factory for Tiled builtin object types. Pass the custom factory to the BasicMap constructor.

    Some shapes accept an anchor argument, which identifies the rect attribute that ia assigned to when obj.position is
    set. They are rectancles, polygons, and poly_lines. This can be customized in a Tiled object by creating a custom
    property anchor with a value of any valid rect attribute. The default is 'topleft'. See geometry.*Geometry class
    constructors for details. For circles and ellipses the anchor is always center.

    NOTE: These sprite objects can be used in the library's spatialash, but they have no image or texture attribute.
    They must be rendered as graphics primitives (see gsdl2's gfx, draw, and gfxdraw modules).

    :param layer: basicmap.BasicLayer; sprites will be added
    :param rectangles: list of rectangle objects
    :param ellipses: list of ellipses objects
    :param polygons: list of polygons objects
    :param poly_lines: list of poly_lines objects
    :param tiles: list of tiles objects
    :return: None
    """
    layer_add = layer.add
    for obj in rectangles:
        # rectangles
        anchor = obj.properties.get('anchor', 'topleft')
        if obj.rotation % 360.0 != 0.0:
            layer_add(Polygon(obj, anchor=anchor))
        else:
            layer_add(Rectangle(obj, anchor=anchor))
    for obj in ellipses:
        # ellipses
        if obj.width == obj.height:
            layer_add(Circle(obj))
        else:
            layer_add(Ellipse(obj))
    for obj in polygons:
        # polygons
        anchor = obj.properties.get('anchor', 'topleft')
        layer_add(Polygon(obj, anchor=anchor))
    for obj in poly_lines:
        # poly_lines
        anchor = obj.properties.get('anchor', 'topleft')
        layer_add(Polygon(obj, closed=False, anchor=anchor))
    for obj in tiles:
        # tiles
        # TODO: how is this compared to a tile in a tilelayer?
        pass


def render_shape(renderer, obj, deltax=0, deltay=0, aa=False, filled=False):
    """render an object using its primitive

    :param renderer: renderer.Renderer used to draw the shape
    :param obj: one of the tiledutils shape classes
    :param deltax: number; move x coordinates before rendering
    :param deltay: number; move y coordinates before rendering
    :return: None
    """
    if isinstance(obj, Rectangle):
        # Rectangle
        if filled:
            points = [(x + deltax, y + deltay) for x, y in obj.points]
            gfx.filled_polygon(renderer, points, obj.color)
        else:
            rect = obj.rect.move(deltax, deltay)
            gfx.rectangle(renderer, rect, obj.color)
    elif isinstance(obj, Circle):
        # Circle
        x, y = obj.origin
        render_circle = gfx.filled_circle if filled else gfx.aacircle if aa else gfx.circle
        render_circle(renderer, x + deltax, y + deltay, obj.radius, obj.color)
    elif isinstance(obj, Ellipse):
        # Ellipse
        x, y = obj.origin
        render_ellipse = gfx.filled_ellipse if filled else gfx.aaellipse if aa else gfx.ellipse
        render_ellipse(renderer, x + deltax, y + deltay, obj.radius_x, obj.radius_y, obj.color)
    elif isinstance(obj, Polygon):
        # Polygon
        d = deltax, deltay
        points = [Vec2d(p) + d for p in obj.points]
        if obj.closed:
            render_polygon = gfx.filled_polygon if filled else gfx.aapolygon if aa else gfx.polygon
            render_polygon(renderer, points, obj.color)
        else:
            p1 = points[0]
            render_line = gfx.aaline if aa else gfx.line
            for p2 in points[1:]:
                render_line(renderer, p1, p2, obj.color)
                p1 = p2
