import os
import sys


appdir = os.path.abspath(os.path.dirname(__file__))
os.chdir(appdir)
appdir = os.path.join(appdir, '../gamelib')
if appdir not in sys.path:
    sys.path.insert(0, appdir)

import gsdl2

import basicmap
import context
import pipestdio
from sprite import TextureSprite
import tiledutils
from vec2d import Vec2d


class Settings:
    resolution = ((1024, 768), (1920, 1080))[0]
    display_flags = (0, gsdl2.sdlconstants.SDL_WINDOW_FULLSCREEN)[0]
    scroll_speed = Vec2d(200.0, 200.0)    # pixels per second
    log_stdout = False
    aa_shapes = True
    filled_shapes = True
    map_names = (
        'Gumm no swamps',           # 0: legacy gummmade Tiled map
        '001-1',                    # 1: the mana world
        'Gummnoswamps-1024x1024',   # 2: great big gigantic map to exercise the loader (takes many seconds)
        'basicMap_object_layer',    # 3: DR0ID's eentsy weentsy object map
        'geoms',                    # 4: shows a whopping 50% hit when using AA rendering
    )
    default_map = 4


if Settings.log_stdout:
    pipestdio.stdout.to_logfile('00.log')


def get_bottom(o):
    return o.rect.bottom


class Camera(object):
    def __init__(self):
        self.rect = gsdl2.Rect((0, 0), Settings.resolution)
        self.picker = self.rect.inflate(*Settings.scroll_speed)
        self._pos = Vec2d(self.rect.center)
        self._old_pos = Vec2d(self.rect.center)

    def _get_pos(self):
        return self._pos
    def _set_pos(self, val):
        self._pos[:] = val
    pos = property(_get_pos, _set_pos)

    def _get_old_pos(self):
        return self._old_pos
    def _set_old_pos(self, val):
        self._old_pos[:] = val
    old_pos = property(_get_old_pos, _set_old_pos)

    def update(self, vxy):
        self.old_pos = self.pos
        self.pos += vxy
        self.picker.center = self.pos


class Game(context.Context):
    def __init__(self):
        super(Game, self).__init__(gsdl2.time.FixedDriver(self.update, 1.0 / 30.0))
        self.draw_sched = self.clock.new_schedule(self.draw, 0.0, keep_history=True)
        self.clock.new_schedule(self.update_caption, 1.0)
        map_name = Settings.map_names[Settings.default_map]
        self.game_map = basicmap.BasicMap(
            map_name=map_name, map_dir='../data/map', file_name='{}.json'.format(map_name),
            object_factory=tiledutils.object_factory)
        try:
            self.game_map.layers_by_name['Collision'].visible = False
        except:
            pass
        self.camera = Camera()
        self.picked = []
        self.num_picked = 0
        self.movex = 0
        self.movey = 0

    def update(self, dt):
        self.get_events()
        self.camera.update((self.movex, self.movey) * (Settings.scroll_speed * dt))
        self.update_picked()

    def update_picked(self):
        del self.picked[:]
        # del self.num_picked[:]
        for layer in self.game_map.layers:
            if layer.visible:
                picked = sorted(layer.intersect_objects(self.camera.picker), key=get_bottom)
                self.picked.extend(picked)
        self.num_picked = len(self.picked)

    def update_caption(self, dt):
        if Settings.display_flags & gsdl2.sdlconstants.SDL_WINDOW_FULLSCREEN:
            print('{} fps | picked: {} | Filled shapes {} | AA shapes: {}'.format(
                int(self.draw_sched.per_second()), self.num_picked, Settings.filled_shapes, Settings.aa_shapes))
        else:
            gsdl2.display.set_caption('{} fps | picked: {} | Filled shapes {} | AA shapes: {}'.format(
                int(self.draw_sched.per_second()), self.num_picked, Settings.filled_shapes, Settings.aa_shapes))

    def draw(self, sched):
        renderer = gsdl2.display.get_renderer()
        # renderer.set_target()  # debugging artifact; leave the clue for now
        renderer.clear()
        renderer_copy = renderer.copy
        render_shape = tiledutils.render_shape
        dx, dy = self.compute_tween(sched.interp)
        for obj in self.picked:
            if isinstance(obj, TextureSprite):
                renderer_copy(obj.texture, obj.rect.move(-dx, -dy), obj.src_rect)
            elif getattr(obj, 'visible', True):
                # render Tiled object
                render_shape(renderer, obj, -dx, -dy, aa=Settings.aa_shapes, filled=Settings.filled_shapes)
        renderer.present()

    def compute_tween(self, interp):
        dx = dy = 0
        try:
            camera = self.camera
            fm = camera.old_pos
            to = camera.pos
            cam_rect = camera.rect
            cx, cy = fm
            if fm != to:
                cx = fm.x + (to.x - fm.x) * interp
                cy = fm.y + (to.y - fm.y) * interp
            cam_rect.center = cx, cy
            dx = round(cx - cam_rect.w // 2)
            dy = round(cy - cam_rect.h // 2)
        except AttributeError:
            # non-fatal: self is not maintaining the necessary attributes for interpolated rendering
            pass
        return dx, dy

    def on_key_down(self, e):
        if e.scancode == gsdl2.S_ESCAPE:
            self.on_quit()
        elif e.scancode == gsdl2.S_RIGHT:
            self.movex += 1
        elif e.scancode == gsdl2.S_LEFT:
            self.movex -= 1
        elif e.scancode == gsdl2.S_DOWN:
            self.movey += 1
        elif e.scancode == gsdl2.S_UP:
            self.movey -= 1
        elif e.scancode == gsdl2.S_A:
            Settings.aa_shapes = not Settings.aa_shapes
        elif e.scancode == gsdl2.S_F:
            Settings.filled_shapes = not Settings.filled_shapes
        elif e.scancode == gsdl2.S_O:
            try:
                layer = self.game_map.layers_by_name['Objects']
                layer.visible = not layer.visible
            except:
                pass
        elif e.scancode == gsdl2.S_C:
            try:
                layer = self.game_map.layers_by_name['Collision']
                layer.visible = not layer.visible
            except:
                pass

    def on_key_up(self, e):
        if e.scancode == gsdl2.S_RIGHT:
            self.movex -= 1
        elif e.scancode == gsdl2.S_LEFT:
            self.movex += 1
        elif e.scancode == gsdl2.S_DOWN:
            self.movey -= 1
        elif e.scancode == gsdl2.S_UP:
            self.movey += 1

    def on_quit(self):
        self.pop()


def main():
    gsdl2.init()
    gsdl2.display.set_mode(Settings.resolution, Settings.display_flags)
    game = Game()
    context.run(game)


if __name__ == '__main__':
    if '-profile' in sys.argv:
        import cProfile
        import pstats
        cProfile.runctx('main()', globals(), locals(), 'prof.dat')
        p = pstats.Stats('prof.dat')
        p.sort_stats('cumulative').print_stats()
    else:
        main()
