# -*- coding: utf-8 -*-
#
# New BSD license
#
# Copyright (c) DR0ID
# This file is part of HG_pyknic
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the <organization> nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL DR0ID BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
TODO: module description


Versioning scheme based on: http://en.wikipedia.org/wiki/Versioning#Designating_development_stage

::

      +-- api change, probably incompatible with older versions
      |     +-- enhancements but no api change
      |     |
    major.minor[.build[.revision]]
                   |        |
                   |        +-|* x for x bugfixes
                   |
                   +-|* 0 for alpha (status)
                     |* 1 for beta (status)
                     |* 2 for release candidate
                     |* 3 for (public) release

.. versionchanged:: 0.0.0.0
    initial version

"""
from __future__ import print_function

from collections import defaultdict

try:
    import sys
    import traceback
    import json
    import math

    import pygame

    __version__ = '1.0.0.0'

    # for easy comparison as in sys.version_info but digits only
    __version_info__ = tuple([int(d) for d in __version__.split('.')])

    __author__ = "DR0ID"
    __email__ = "dr0iddr0id {at} gmail [dot] com"
    __copyright__ = "DR0ID @ 2016"
    __credits__ = ["DR0ID"]  # list of contributors
    __maintainer__ = "DR0ID"
    __license__ = "New BSD license"

    __all__ = []  # list of public visible parts of this module 

    SCREENSIZE = 1024, 768

    import os
    import sys

    appdir = os.path.abspath(os.path.dirname(__file__))
    os.chdir(appdir)
    appdir = os.path.join(appdir, '../gamelib')
    if appdir not in sys.path:
        sys.path.insert(0, appdir)

    class Sprite(object):

        def __init__(self, x, y, image, layer, orig_y):
            self.image = image
            self.rect = pygame.Rect((x, y), image.get_size())
            self.pos = (x, y)
            self.layer = layer
            self.orig_y = orig_y


    def create_sprite_layers(map_info):
        sprite_layers = []  # sprite_layers[0][y][x]
        for idx_layer, layer in enumerate(map_info.layers):
            if layer.layer_type == "tilelayer":
                if layer.name == 'Collision':
                    continue
                sprite_layers.append([])
                for idx_y, row in enumerate(layer.data_yx):
                    for lti in row:
                        if lti is not None:
                            image = lti.tile_info.image
                            r = image.get_rect()
                            if r.w > 32 or r.h > 32:
                                pygame.draw.rect(image, (255, 0, 0), r, 2)
                            sprite = Sprite(lti.tile_x - 400, lti.tile_y + lti.offset_y - 400, image, idx_layer, lti.tile_y)
                            sprite_layers[idx_layer].append(sprite)
        return sprite_layers


    def load_images(map_info, img_cache=None):
        if img_cache is None:
            img_cache = {}
        for tile in map_info.tiles.values():
            tileset = tile.tileset
            image_name = tileset.image_name

            # load source image first if not already cached
            image_source = img_cache.get(image_name, None)
            if image_source is None:
                # image_source = pygame.image.load("../data/image/tmw/" + image_name)
                print(os.getcwd())
                image_source = pygame.image.load(os.path.abspath("../data/map/" + image_name))
                if tileset.transparent_color:
                    transparent_color = pygame.Color(tileset.transparent_color)
                    image_source.set_colorkey(transparent_color, pygame.RLEACCEL)
                    image_source = image_source.convert()
                else:
                    image_source = image_source.convert_alpha()
                img_cache[image_name] = image_source

            # load tile image from source image
            key = (image_name, tile.spritesheet_x, tile.spritesheet_y, tile.angle, tile.flip_x, tile.flip_y)
            tile_image = img_cache.get(key, None)
            if tile_image is None:
                tile_image = pygame.Surface((tileset.tile_width, tileset.tile_height), pygame.SRCALPHA)
                # tile_image.fill((0, 0, 0, 0), None, pygame.BLEND_RGBA_MULT)
                source_rect = pygame.Rect(tile.spritesheet_x, tile.spritesheet_y, tileset.tile_width, tileset.tile_height)
                tile_image.blit(image_source, (0, 0), source_rect)
                # important: order of operations matter: rotate first and then flip the image
                if tile.angle > 0:
                    tile_image = pygame.transform.rotate(tile_image, -tile.angle)
                if tile.flip_x or tile.flip_y:
                    tile_image = pygame.transform.flip(tile_image, tile.flip_x, tile.flip_y)
                img_cache[key] = tile_image
            tile.image = tile_image


    def load_graph(map_info):
        graph = Graph(NavNode, NavEdge)
        # add nodes
        _nodes = defaultdict(dict)  # {layer_idx : {(tile_x, tile_y): cur_node}}
        for idx_layer, layer in enumerate(map_info.layers):
            if not layer.name.startswith("graph"):
                continue
            for idx_y, row in enumerate(layer.data_yx):
                for idx_x, layer_tile in enumerate(row):
                    if layer_tile is None:
                        pass
                    else:
                        tile = layer_tile.tile_info
                        if 'edges' not in tile.properties:
                            raise KeyError("edges should be in a cur_node tile")
                        angle = layer_tile.tile_info.angle
                        if angle != 0:
                            raise ArithmeticError("no rotated graph tiles allowed!")
                        edges = set([property_mapping[_edge.strip()] for _edge in tile.properties['edges'].split(',')])
                        edge_properties = dict([(property_mapping[_key], _val) for _key, _val in tile.properties.items()
                                                if _key in property_mapping.keys()])
                        wx = layer_tile.tile_x + layer.tile_w // 2
                        wy = layer_tile.tile_y + layer.tile_h // 2
                        cur_node = NavNode(wx, wy, layer_tile.tile_x, layer_tile.tile_y, edges, edge_properties,
                                           idx_layer)
                        _nodes[idx_layer][(idx_x, idx_y)] = cur_node
                        if abs(idx_x - layer_tile.tile_x // 64) > 1 or abs(idx_y - layer_tile.tile_y // 64) > 0:
                            print(layer_tile)
                        graph.add_node(cur_node)

        # add edges
        neighbour_coordinates = (
            (1, 0, e_i, e_o), (1, 1, se_i, se_o), (0, 1, s_i, s_o), (-1, 1, sw_i, sw_o), (-1, 0, w_i, w_o),
            (-1, -1, nw_i, nw_o), (0, -1, n_i, n_o), (1, -1, ne_i, ne_o)
        )
        layer_indices = sorted(_nodes.keys())
        for layer_idx_idx, layer_idx in enumerate(layer_indices):
            # for key, _orig_cur_node in _nodes[layer_idx].items():
            sorted_keys = sorted(_nodes[layer_idx].keys())
            for key in sorted_keys:
                _orig_cur_node = _nodes[layer_idx][key]
                idx_x, idx_y = key
                for dx, dy, d_i, d_o in neighbour_coordinates:
                    cur_x = idx_x + dx
                    cur_y = idx_y + dy

                    cur_d = (d_i, d_o)
                    relevant_edge_props = dict([(_edge_key, _v) for _edge_key, _v in cur_node.edge_properties.items()
                                                if _edge_key in cur_d])
                    for d_d in cur_d:
                        neighbour_node = _nodes[layer_idx].get((cur_x, cur_y), None)
                        cur_node = _orig_cur_node
                        if neighbour_node is None:
                            # check for special node properties
                            prop_value = relevant_edge_props.get(d_d, "")
                            # check for layer change property
                            if prop_value.startswith("layer"):
                                if d_d in edge_in_set and prop_value.endswith('++'):
                                    raise NotImplemented("at the moment only upper layers are supported for in edges")
                                if d_d in edge_out_set and prop_value.endswith('--'):
                                    raise NotImplemented("at the moment only upper layers are supported for out edges")
                                # find potentially connecting node on other layer
                                for upper_layer_idx in layer_indices[layer_idx_idx + 1:]:
                                    neighbour_node = _nodes[upper_layer_idx].get((cur_x, cur_y), None)
                                    if neighbour_node is not None:
                                        break

                        # if it is still None (none found in other layers) continue
                        if neighbour_node is None:
                            continue

                        # swap nodes to make sure that the right edge is inserted
                        # the swap of the edge direction is needed too
                        if d_d in edge_in_set:
                            neighbour_node, cur_node = cur_node, neighbour_node
                            d_d = connectors[d_d]

                        # treat all edges as if they where a in edge on the node (e.g. neighbour_node -> node)
                        if d_d in cur_node.edges and connectors[d_d] in neighbour_node.edges \
                                and not graph.is_edge_present(cur_node.id, neighbour_node.id):
                            layer_change = neighbour_node.layer_index - cur_node.layer_index
                            edge = NavEdge(d_d, layer_change, cur_node.id, neighbour_node.id)
                            if cur_node.id == neighbour_node.id:
                                raise "same nodes {0}".format(cur_node.id)
                            graph.add_edge(edge)

        return graph


    def draw_graph(screen, graph):
        # return
        for from_n_id, edge_dict in graph.edges.items():
            for to_n_id, edge in edge_dict.items():
                from_node = graph.nodes[from_n_id]
                to_node = graph.nodes[to_n_id]
                pygame.draw.line(screen, (255, 0, 0), (from_node.world_x, from_node.world_y),
                                 (to_node.world_x, to_node.world_y), 3)
        for node in sorted(graph.nodes.values(), key=lambda a: a.layer_index):
            pygame.draw.circle(screen, (255 - 80 * node.layer_index, 255, 80 * node.layer_index, 50),
                               (node.world_x, node.world_y), 12 - 3 * node.layer_index)


    def main():

        # json_map_data = json.load(open("tunnel2.json"))
        # json_map_data = json.load(open("../data/map/991-1.json"))
        json_map_data = json.load(open("../data/map/Gummnoswamps-1024x1024.json"))
        # map_info = load_map_info(json_map_data)  # MapInfo -> layers:[LayerInfo, ...], tilesets: {gid: TileInfo}
        import pytmxloader
        map_info = pytmxloader.pytmxloader.load_map_from_data(json_map_data)  # MapInfo -> layers:[LayerInfo, ...], tilesets: {gid: TileInfo}

        pygame.init()
        # screen = pygame.display.set_mode(SCREENSIZE, pygame.RESIZABLE | pygame.SRCALPHA, 32)
        screen = pygame.display.set_mode(SCREENSIZE, pygame.SRCALPHA)
        camera = screen.get_rect()

        # load images and create sprites
        _img_cache = {}  # {key: img}
        load_images(map_info, _img_cache)

        # graph = load_graph(map_info)

        sprite_layers = create_sprite_layers(map_info)
        layer_visibility = [_layer.visible for _layer in map_info.layers if _layer.layer_type == "tilelayer"]
        sprites = [_spr for _layer_idx, _layer in enumerate(sprite_layers) for _spr in _layer if
                   layer_visibility[_layer_idx]]
        current_layer_idx = 0

        # hero
        # start_node = graph.nodes.values()[-1]
        # hero_sprite = Sprite(start_node.world_x, start_node.world_y, pygame.Surface((15, 15), pygame.SRCALPHA), start_node.layer_index + 0.5 - 1.0)
        # sprites.append(hero_sprite)
        vel = 80
        dx = 0
        dy = 0
        # current_node = start_node
        current_edge = None

        hero_mode = 1

        def sort_key(_s): return _s.layer, _s.pos[0], _s.orig_y
        sprites.sort(key=sort_key)
        # angles = {e_o: 0, n_o: 90, w_o: 180, s_o: 270}

        clock = pygame.time.Clock()
        screen_blit = screen.blit
        running = True
        while running:
            # events
            for event in pygame.event.get():

                if event.type == pygame.QUIT:
                    running = False
                elif event.type == pygame.KEYDOWN:
                    if event.key == pygame.K_ESCAPE:
                        running = False
                    elif event.key == pygame.K_SPACE:
                        current_layer_idx += 1
                        current_layer_idx %= len(layer_visibility)
                    elif event.key == pygame.K_v:
                        layer_visibility[current_layer_idx] = not layer_visibility[current_layer_idx]
                        if layer_visibility[current_layer_idx]:
                            # is visible, add sprites
                            sprites.extend(sprite_layers[current_layer_idx])
                        else:
                            # is invisible now, remove sprites
                            for spr in sprite_layers[current_layer_idx]:
                                sprites.remove(spr)
                        sprites.sort(key=sort_key)
                    elif event.key == pygame.K_m:
                        hero_mode += 1
                        hero_mode %= 3
                    elif event.key == pygame.K_a:
                        hero_sprite.layer -= 1
                        sprites.sort(key=sort_key)
                    elif event.key == pygame.K_q:
                        hero_sprite.layer += 1
                        sprites.sort(key=sort_key)
                    elif hero_mode == 0:
                        dx = 0
                        dy = 0
                        if event.key == pygame.K_DOWN:
                            dy += 1
                        elif event.key == pygame.K_UP:
                            dy -= 1
                        elif event.key == pygame.K_LEFT:
                            dx -= 1
                        elif event.key == pygame.K_RIGHT:
                            dx += 1
                    elif hero_mode >= 1:
                        key_direction = None
                        if event.key == pygame.K_DOWN:
                            key_direction = s_o
                        elif event.key == pygame.K_UP:
                            key_direction = n_o
                        elif event.key == pygame.K_LEFT:
                            key_direction = w_o
                        elif event.key == pygame.K_RIGHT:
                            key_direction = e_o
                        if key_direction:
                            nx = current_node.world_x - hx
                            ny = current_node.world_y - hy
                            n_dist_sq = nx * nx + ny * ny
                            if n_dist_sq < 32 * 32:

                                if key_direction in current_node.edges:
                                    node_edges = graph.edges[current_node.id]
                                    for to_id, edge in node_edges.items():
                                        if edge.direction == key_direction:
                                            current_node = graph.nodes[to_id]
                                            current_edge = edge
                                            if edge.layer_change != 0:
                                                hero_sprite.layer += edge.layer_change
                                                sprites.sort(key=sort_key)
                                            break

                elif event.type == pygame.KEYUP:
                    if event.key == pygame.K_DOWN:
                        dy -= 1
                    elif event.key == pygame.K_UP:
                        dy += 1
                    elif event.key == pygame.K_LEFT:
                        dx += 1
                    elif event.key == pygame.K_RIGHT:
                        dx -= 1

            delta_t = clock.tick() / 1000.0  # seconds
            current_time = pygame.time.get_ticks() / 1000.0  # seconds
            pulse_value = math.sin(current_time * 2.0 * math.pi / 0.5)
            print(clock.get_fps())

            for spr in [s for s in sprites if camera.colliderect(s.rect)]:
                screen_blit(spr.image, spr.pos)


            pygame.display.flip()
            screen.fill((255, 0, 255))

        pygame.quit()


    if __name__ == '__main__':
        if '-profile' in sys.argv:
            import cProfile
            import pstats
            cProfile.runctx('main()', globals(), locals(), 'prof.dat')
            p = pstats.Stats('prof.dat')
            p.sort_stats('cumulative').print_stats()
        else:
            main()

except Exception as ex:
    # trying to catch any exceptions and print them out in a way that the user will see them!
    sys.stderr.write("Error:")
    sys.stderr.write('\n')
    sys.stderr.write('\n')
    sys.stderr.write(str(ex.__class__.__name__))
    sys.stderr.write(str(ex))
    sys.stderr.write('\n')
    sys.stderr.write('\n')
    exc_type, exc_value, exc_traceback = sys.exc_info()
    traceback.print_tb(exc_traceback, file=sys.stderr)
    sys.stderr.write('\n')
    sys.stderr.write('Press any key... ')
    sys.stderr.flush()
    # sys.stdin.readline()
