#!/usr/bin/env python

# This file is part of Gummworld2.
#
# Gummworld2 is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Gummworld2 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with Gummworld2.  If not, see <http://www.gnu.org/licenses/>.

# Compatible: Python 2.7, Python 3.2

"""14_collision_sandbox.py - A sandbox / integration test for the
geometry module's collision routines in Gummworld2.

This test uses the eyeballs to validate. Do the following:
    
    1.  Ignore the lines for now.
    2.  Drag one small rect, poly, and circle completely into the other large
        2D shapes--one each. No sides should be touching
    3.  Observe that the shapes turn red as their edges cross.
    4.  Observe that the shapes stay red when one is completely inside another.
    5.  Press a key to swith between forward and reverse order collision tests.
    6.  Repeat steps 2-5 two times, rotating shape combinations.
    7.  Cross the lines and observe they turn red.
    8.  Drag a line completely into one of the 2D shapes. No sides should be
        touching.
    9.  Observe that the shapes turn red as their edges cross.
    10. Observe that the shapes stay red when one is completely inside another.
    11. Press a key to swith between forward and reverse order collision tests.
    12. Repeat steps 8-11 two times, rotating shape combinations.
"""

__version__ = '$Id: 14_collision_sandbox.py 407 2013-08-12 15:11:30Z stabbingfinger@gmail.com $'
__author__ = 'Gummbum, (c) 2011-2014'


import os
import sys


appdir = os.path.abspath(os.path.dirname(__file__))
os.chdir(appdir)
appdir = '/'.join((appdir, '../gamelib'))
if appdir not in sys.path:
    sys.path.insert(0, appdir)

import gsdl2
from gsdl2 import Color, Rect, MOUSEMOTION, MOUSEBUTTONDOWN, MOUSEBUTTONUP, KEYDOWN, QUIT, S_ESCAPE, S_SPACE

import geometry


class Tabs(object):
    def on_tab(self, p):
        for t in self.tabs:
            if t.collidepoint(p):
                self._dirty = True
                return t


class RectGeom(geometry.RectGeometry, Tabs):
    def __init__(self, x, y, w, h):
        geometry.RectGeometry.__init__(self, x, y, w, h)
        r = self.rect
        self.tabs = [
            Rect(0, 0, 7, 7),
        ]
        self.tabs[0].center = r.center
        self.prev = [t.center for t in self.tabs]
        self.color = None

    def update(self):
        self.rect.center = self.tabs[0].center


class LineGeom(geometry.LineGeometry, Tabs):

    def __init__(self, x1, y1, x2, y2):
        geometry.LineGeometry.__init__(self, x1, y1, x2, y2)
        self.tabs = [
            Rect(0, 0, 7, 7),
            Rect(0, 0, 7, 7),
        ]
        self.tabs[0].center = self.end_points[0]
        self.tabs[1].center = self.end_points[1]
        self.prev = [t.center for t in self.tabs]
        self.color = None

    def update(self):
        self.points = self.tabs[0].center, self.tabs[1].center


class PolyGeom(geometry.PolyGeometry, Tabs):

    def __init__(self, points):
        geometry.PolyGeometry.__init__(self, points)
        self.tabs = [Rect(0, 0, 7, 7) for p in self.points]
        for i, p in enumerate(self.points):
            self.tabs[i].center = p
        self.prev = [t.center for t in self.tabs]
        self.color = None

    def update(self):
        now = [t.center for t in self.tabs]
        if self.prev != now:
            self.points = now
            self.prev = now


class CircleGeom(geometry.CircleGeometry, Tabs):

    def __init__(self, origin, radius):
        geometry.CircleGeometry.__init__(self, origin, radius)
        self.tabs = [
            Rect(0, 0, 7, 7),
        ]
        self.tabs[0].center = self.origin
        self.prev = [t.center for t in self.tabs]
        self.color = None

    def update(self):
        self.position = self.tabs[0].center


class EllipseGeom(geometry.EllipseGeometry, Tabs):

    def __init__(self, origin, radius_x, radius_y):
        geometry.EllipseGeometry.__init__(self, origin, radius_x, radius_y)
        self.tabs = [
            Rect(0, 0, 7, 7),
        ]
        self.tabs[0].center = self.origin
        self.prev = [t.center for t in self.tabs]
        self.color = None

    def update(self):
        self.origin = self.tabs[0].center


def check_collisions(sprites):
    for s in sprites:
        s.color = Color('green')
    for i, s1 in enumerate(sprites):
        for j, s2 in enumerate(sprites[i:]):
            if j > 0:
                if s1.collided(s1, s2):
                    s1.color = Color('red')
                    s2.color = Color('red')


def update_sprites(sprites):
    for s in sprites:
        s.update()


def draw_sprites(renderer, sprites):
    for s in sprites:
        if isinstance(s, RectGeom):
            draw_rect(renderer, s.rect, s.color)
        elif isinstance(s, LineGeom):
            draw_line(renderer, s.p1, s.p2, s.color)
        elif isinstance(s, PolyGeom):
            draw_poly(renderer, s.points, s.color)
        elif isinstance(s, CircleGeom):
            x, y = s.origin
            draw_circle(renderer, x, y, s.radius, s.color)
        elif isinstance(s, EllipseGeom):
            x, y = s.origin
            draw_ellipse(renderer, x, y, s.radius_x, s.radius_y, s.color)
        else:
            raise(ValueError, 'bad sprite ' + str(s))
        for t in s.tabs:
            draw_rect(renderer, t, Color('yellow'))


gsdl2.init()
screen = gsdl2.display.set_mode((400, 400))
renderer = gsdl2.display.get_renderer()
gsdl2.display.set_caption('Space: Test sprites reversed')
screen_rect = screen.get_rect()
clock = gsdl2.time.Clock()

rect1 = RectGeom(0, 0, 40, 40)
rect2 = RectGeom(40, 40, 80, 80)
line1 = LineGeom(40, 170, 90, 210)
line2 = LineGeom(20, 200, 80, 250)
poly1 = PolyGeom([(202, 20), (202, 62), (242, 42)])
poly2 = PolyGeom([(202, 100), (240, 150), (235, 200), (150, 250)])
circle1 = CircleGeom((225, 300), 20)
circle2 = CircleGeom((300, 300), 50)
ellipse1 = EllipseGeom((350, 40), 20, 11)
ellipse2 = EllipseGeom((320, 130), 60, 41)

sprites_fwd = (rect1, rect2, line1, line2, poly1, poly2, circle1, circle2, ellipse1, ellipse2)
sprites_rev = list(reversed(sprites_fwd))
sprites = sprites_fwd

draw_circle = gsdl2.gfx.circle
draw_line = gsdl2.gfx.line
draw_rect = gsdl2.gfx.rectangle
draw_poly = gsdl2.gfx.polygon
draw_ellipse = gsdl2.gfx.ellipse

selected = None

while 1:
    
    clock.tick(60)
    
    for e in gsdl2.event.get():
        if e.type == MOUSEMOTION:
            if selected:
                selected.center = e.pos
        elif e.type == MOUSEBUTTONDOWN:
            for s in sprites:
                selected = s.on_tab(e.pos)
                if selected:
                    break
        elif e.type == MOUSEBUTTONUP:
            selected = None
        elif e.type == KEYDOWN:
            if e.scancode == S_ESCAPE:
                quit()
            elif e.scancode == S_SPACE:
                if sprites is sprites_fwd:
                    sprites = sprites_rev
                    gsdl2.display.set_caption('Space: Test sprites forward')
                else:
                    sprites = sprites_fwd
                    gsdl2.display.set_caption('Space: Test sprites reversed')
        elif e.type == QUIT:
            quit()

    update_sprites(sprites)
    check_collisions(sprites)

    renderer.clear()
    draw_sprites(renderer, sprites)
    renderer.present()
